class ApexproError(Exception):
    """Base error class for all exceptions raised in this library.
    Will never be raised naked; more specific subclasses of this exception will
    be raised when appropriate."""


class ApexproApiError(ApexproError):

    def __init__(self, response):
        self.status_code = response.status_code
        try:
            self.msg = response.json()
        except ValueError:
            self.msg = response.text
        self.response = response
        self.request = getattr(response, 'request', None)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'ApexproApiError(status_code={}, response={})'.format(
            self.status_code,
            self.msg,
        )


class TransactionReverted(ApexproError):

    def __init__(self, tx_receipt):
        self.tx_receipt = tx_receipt






import hashlib
import json
import time
import urllib

import zklink_sdk as sdk
from apexpro.constants import URL_SUFFIX, APEX_OMNI_HTTP_TEST
from apexpro.helpers.request_helpers import random_client_id
from apexpro.http_private import HttpPrivate


class HttpPrivate_v3(HttpPrivate):

    def generate_nonce_v3(self, l2Key, ethAddress, chainId,
                       refresh="false"
                       ):
        """"
        POST: Generate nonce.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-generate-nonce
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/generate-nonce"
        return self._private_request(
            method="POST",
            path=path,
            data={
                'l2Key': l2Key,
                'ethAddress': ethAddress,
                'chainId': chainId,
                'category': 'CATEGORY_API',
                'refresh': refresh,
            }
        )

    def register_user_v3(
            self,
            nonce,
            l2Key=None,
            seeds=None,
            ethereum_address=None,
            referred_by_affiliate_link=None,
            country=None,
            isLpAccount=None,
            eth_mul_address=None,
            sourceFlag=None,
    ):
        """"
        POST Registration & Onboarding.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-registration-amp-onboarding
        :returns: Request results as dictionary.
        """
        l2_key = l2Key or self.zk_l2Key
        l2_seeds = seeds or self.zk_seeds

        if l2_key is None:
            raise ValueError(
                'zk l2Key is required'
            )

        if l2_seeds is None:
            raise ValueError(
                'zk seeds is required'
            )

        eth_address = ethereum_address or self.default_address

        msg = str(l2_key.removeprefix('0x') + eth_address + nonce).lower()
        message = hashlib.sha256()
        message.update(msg.encode())  # Encode as UTF-8.
        msgHash = message.digest()


        EC_ORDER = '3618502788666131213697322783095070105526743751716087489154079457884512865583';

        bn1 = int(msgHash.hex(), 16)
        bn2 = int (EC_ORDER, 10)

        bn3 = hex(bn1)
        signMsg = hex(bn1.__mod__(bn2))

        seeds = bytes.fromhex(l2_seeds)
        signerSeed = sdk.ZkLinkSigner.new_from_seed(seeds)
        signatureOnboard = signerSeed.sign_musig(signMsg.removeprefix('0x').encode())

        apiKeyHash = hex(bn1).removeprefix('0x') + '|' + signMsg.removeprefix('0x')

        path = URL_SUFFIX + "/v3/new-onboarding"
        onboardingRes = self._private_request(
            method="POST",
            path=path,
            data= {
                'l2Key': l2_key,
                'referredByAffiliateLink': referred_by_affiliate_link,
                'ethereumAddress': eth_address,
                'country': country,
                'category': 'CATEGORY_API',
                'isLpAccount': isLpAccount,
                'ethMulAddress': eth_mul_address,
                'sourceFlag': sourceFlag,
                'apiKeyHash': apiKeyHash,
            },
            headers={
                'APEX-SIGNATURE': signatureOnboard.signature,
                'APEX-ETHEREUM-ADDRESS': eth_address,
            }
        )
        if onboardingRes.get('data') is not None:
            self.user = onboardingRes.get('data').get('user')
            self.accountV3 = onboardingRes.get('data').get('account')
            key = onboardingRes['data']['apiKey']['key']
            secret = onboardingRes['data']['apiKey']['secret']
            passphrase = onboardingRes['data']['apiKey']['passphrase']
            self.api_key_credentials = {'key': key,'secret': secret, 'passphrase': passphrase}
        return onboardingRes


    def derive_zk_key(
            self,
            ethereum_address=None,
    ):
        msgHeader = 'ApeX Omni Mainnet'
        if self.endpoint == APEX_OMNI_HTTP_TEST:
            msgHeader = 'ApeX Omni Testnet'
        signature = self.starkeySigner.sign_zk_message(
            ethereum_address or self.default_address,
            msgHeader,
            )

        seedstr = signature.removeprefix("0x")
        seeds = bytes.fromhex(seedstr)
        self.zk_seeds = seedstr
        signerSeed = sdk.ZkLinkSigner.new_from_seed(seeds)
        pubKey = signerSeed.public_key()
        self.zk_l2Key = pubKey
        pubKeyHash = sdk.get_public_key_hash(pubKey)

        return {
            'seeds': seedstr,
            'l2Key': pubKey,
            'pubKeyHash': pubKeyHash
        }

    def user_v3(self, **kwargs):
        """"
        GET Retrieve User Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/user"
        userRes = self._get(
            endpoint=path,
            params=kwargs
        )
        self.user = userRes.get('data')
        return userRes

    def get_user_v3(self, **kwargs):
        """"
        GET Retrieve User Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/user"
        userRes = self._get(
            endpoint=path,
            params=kwargs
        )
        self.user = userRes.get('data')
        return userRes


    def modify_user_v3(self, **kwargs):
        """"
        POST Edit User Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-edit-user-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/modify-user"

        return self._post(
            endpoint=path,
            data=kwargs
        )


    def get_account_v3(self, **kwargs):
        """"
        GET Retrieve User Account Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-account-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/account"
        accountRes =  self._get(
            endpoint=path,
            params=kwargs
        )
        self.accountV3 = accountRes.get('data')
        self.default_address = self.accountV3.get('ethereumAddress')
        return accountRes.get('data')

    def transfers_v3(self, **kwargs):
        """"
        GET Retrieve User Deposit Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-deposit-data
        :returns: Request results as dictionary.
        """

        kwargs.update(subAccountId='0')
        path = URL_SUFFIX + "/v3/transfers"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def transfer_v3(self, **kwargs):
        """"
        GET Retrieve User Deposit Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-deposit-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/transfer"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def contract_transfers_v3(self, **kwargs):
        """"
        GET Retrieve User Deposit Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-deposit-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/contract-transfers"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def contract_transfer_v3(self, **kwargs):
        """"
        GET Retrieve User Deposit Data.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-deposit-data
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/contract-transfer"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def withdraw_list_v3(self, **kwargs):
        """"
        GET Retrieve User Withdrawal List.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-withdrawal-list
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/withdraw-list"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def contract_transfer_limit_v3(self, **kwargs):
        """"
        GET Retrieve Withdrawal & Transfer Limits.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-withdrawal-amp-transfer-limits
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/contract-transfer-limit"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def fills_v3(self, **kwargs):
        """"
        GET Retrieve Trade History.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-trade-history
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/fills"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def order_fills_v3(self, **kwargs):

        path = URL_SUFFIX + "/v3/order-fills"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def delete_order_v3(self, **kwargs):
        """"
        POST Cancel Order.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-cancel-order
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/delete-order"
        return self._post(
            endpoint=path,
            data=kwargs
        )


    def delete_order_by_client_order_id_v3(self, **kwargs):
        """"
        POST Cancel Order.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-cancel-order
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/delete-client-order-id"
        return self._post(
            endpoint=path,
            data=kwargs
        )

    def delete_open_orders_v3(self, **kwargs):
        """"
        POST Cancel all Open Orders
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-cancel-all-open-orders
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/delete-open-orders"
        return self._post(
            endpoint=path,
            data=kwargs
        )

    def open_orders_v3(self, **kwargs):
        """"
        GET Retrieve Open Orders.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-open-orders
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/open-orders"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def history_orders_v3(self, **kwargs):
        """"
        GET Retrieve All Order History.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-all-order-history
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/history-orders"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def get_order_v3(self, **kwargs):
        """"
        GET Retrieve Order ID.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-order-id
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/order"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def get_order_by_client_order_id_v3(self, **kwargs):
        """"
        GET Retrieve Order ID.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-order-id
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/order-by-client-id"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def funding_v3(self, **kwargs):
        """"
        GET Retrieve Funding Rate.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-funding-rate
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/funding"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def historical_pnl_v3(self, **kwargs):
        """"
        GET Retrieve User Historial Profit and Loss.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-user-historial-profit-and-loss
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/historical-pnl"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def yesterday_pnl_v3(self, **kwargs):
        """"
        GET Retrieve Yesterday's Profit & Loss.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-yesterday-39-s-profit-amp-loss
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/yesterday-pnl"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def history_value_v3(self, **kwargs):
        """"
        GET Retrieve Historical Asset Value.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-historical-asset-value
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/history-value"
        return self._get(
            endpoint=path,
            params=kwargs
        )


    def get_worst_price_v3(self, **kwargs):
        """"
        get market price from orderbook
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-worst-price
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/get-worst-price"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def get_order_by_client_id_v3(self, **kwargs):
        """"
        get market price from orderbook
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-retrieve-worst-price
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/order-by-client-id"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def get_account_balance_v3(self, **kwargs):
        """"
        get market price from orderbook
        :param kwargs: See
        https://api-docs.pro.apex.exchange/?lang=zh-TW#privateapi-get-retrieve-user-account-balance
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/account-balance"
        return self._get(
            endpoint=path,
            params=kwargs
        )



    def set_initial_margin_rate_v3(self, **kwargs):
        """"
        get market price from orderbook
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-post-sets-the-initial-margin-rate-of-a-contract
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/set-initial-margin-rate"
        return self._post(
            endpoint=path,
            data=kwargs
        )


    def change_pub_key_v3(self,
                          chainId,
                          zkAccountId,
                          seeds,
                          newPkHash,
                          nonce,
                          l2Key,
                          feeToken=None,
                          asset='USDT',
                          subAccountId='0',
                          fee='0',
                          ethPrivateKey = None,
                          ethSignatureType='EthECDSA',
                          signature=None,
                          ethSignature=None,
                          timestamp=None,):
        path = URL_SUFFIX + "/v3/change-pub-key"

        times = timestamp or int(time.time())

        #signer1 = sdk.ZkLinkSigner().new_from_hex_eth_signer(ethPrivateKey)
        #pubKey = signer1.public_key()
        #newPkHash = sdk.get_public_key_hash(pubKey)

        feeToken = feeToken or self.configV3.get('spotConfig').get('global').get('defaultChangePubKeyFeeTokenId')

        builder = sdk.ChangePubKeyBuilder(chainId, int(zkAccountId), int(subAccountId), newPkHash, int(feeToken), fee, int(nonce), ethSignature, int(times))
        tx = sdk.ChangePubKey(builder)


        seedsByte = bytes.fromhex(seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner.new_from_seed(seedsByte)
        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature
        if  ethSignatureType != 'Onchain' and ethPrivateKey is not None:
            signer = sdk.Signer(ethPrivateKey, sdk.L1SignerType.ETH())
            authSigner = json.loads(signer.sign_change_pubkey_with_eth_ecdsa_auth(tx).tx)
            ethSignature = authSigner.get('ethAuthData').get('ethSignature')

        print(builder)
        print(signature)

        return self._post(
            endpoint=path,
            data={
                'chainId': chainId,
                'zkAccountId': zkAccountId,
                'subAccountId' : subAccountId,
                'newPkHash': newPkHash,
                'feeToken': feeToken,
                'nonce':nonce,
                'l2Key': l2Key,
                'fee': fee,
                'ethSignatureType': ethSignatureType,
                'signature': signature,
                'ethSignature': ethSignature,
                'timestamp':times
            }
        )

    def withdraw_fee_v3(self, **kwargs):
        """"
        GET Fast & Cross-Chain Withdrawal Fees.
        :param kwargs: See
        https://api-docs.pro.apex.exchange/#privateapi-get-fast-amp-cross-chain-withdrawal-fees
        :returns: Request results as dictionary.
        """

        path = URL_SUFFIX + "/v3/withdraw-fee"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def all_apikeys_v3(self, **kwargs):

        path = URL_SUFFIX + "/v3/all-api-keys"
        return self._get(
            endpoint=path,
            params=kwargs
        )

    def get_repayment_price_v3(self, repaymentPriceTokens, clientId=None ):

        repaymentPriceTokens = urllib.parse.quote(repaymentPriceTokens)

        clientId = clientId or random_client_id()
        path = URL_SUFFIX + "/v3/repayment-price"
        return self._get(
            endpoint=path,
            params={
                'repaymentPriceTokens':repaymentPriceTokens,
                'clientId': clientId
            }
        )

    def apikey_bind_ip_v3(self, **kwargs):

        path = URL_SUFFIX + "/v3/bind-ip"
        return self._post(
            endpoint=path,
            data=kwargs
        )

    def delete_api_key_v3(self, **kwargs):

        path = URL_SUFFIX + "/v3/delete-api-key"
        return self._post(
            endpoint=path,
            data=kwargs
        )

    def generate_api_key_v3(self, nonce,
                            ips, remark
                            ):
        l2_key = self.zk_l2Key
        l2_seeds = self.zk_seeds

        if l2_key is None:
            raise ValueError(
                'zk l2Key is required'
            )

        if l2_seeds is None:
            raise ValueError(
                'zk seeds is required'
            )

        eth_address =  self.default_address

        msg = str(l2_key.removeprefix('0x') + eth_address + nonce).lower()
        message = hashlib.sha256()
        message.update(msg.encode())  # Encode as UTF-8.
        msgHash = message.digest()


        EC_ORDER = '3618502788666131213697322783095070105526743751716087489154079457884512865583';

        bn1 = int(msgHash.hex(), 16)
        bn2 = int (EC_ORDER, 10)

        bn3 = hex(bn1)
        signMsg = hex(bn1.__mod__(bn2))

        seeds = bytes.fromhex(l2_seeds)
        signerSeed = sdk.ZkLinkSigner.new_from_seed(seeds)
        signatureOnboard = signerSeed.sign_musig(signMsg.removeprefix('0x').encode())

        path = URL_SUFFIX + "/v3/generate-api-key"
        return self._post(
            endpoint=path,
            data={
                'signature':signatureOnboard.signature,
                'l2Key':l2_key,
                'ips':ips,
                'remark':remark,
            }
        )

    def withdraws_by_time_and_status_v3(self, **kwargs):
        path = URL_SUFFIX + "/v3/withdraws-by-time-and-status"
        return self._get(
            endpoint=path,
            params=kwargs
        )



# -*- coding: utf-8 -*-

"""
apexpro
------------------------

apexpro is a lightweight and high-performance API connector for the
RESTful and WebSocket APIs of the Apex pro exchange.

Documentation can be found at
https://api-docs.pro.apex.exchange/#introduction

:copyright: (c) 2020-2022 apexpro-exchange
:license: MIT License

"""

import time
import hmac
import json
import logging
import requests

from web3 import Web3
from apexpro.eth_signing import SignWithWeb3, SignOnboardingAction
from apexpro.eth_signing import SignWithKey
from apexpro.constants import REGISTER_ENVID_MAIN, APEX_HTTP_MAIN, URL_SUFFIX, NETWORKID_MAIN
from apexpro.starkex.helpers import private_key_to_public_key_pair_hex

from datetime import datetime as dt
from concurrent.futures import ThreadPoolExecutor

from .eth import Eth
from .exceptions import FailedRequestError, InvalidRequestError
from .models import configDecoder

try:
    from simplejson.errors import JSONDecodeError
except ImportError:
    from json.decoder import JSONDecodeError

# Versioning.
VERSION = '1.0.0'


class HTTP:
    """
    Connector for Apexpro's HTTP API.

    The below class is maintained for compatibility reasons. You should
    prefer using the market-specific classes maintained in
    inverse_perpetual.py, usdc_perpetual.py, spot.py, etc

    :param endpoint: The endpoint URL of the HTTP API, e.g.
        'https://dev.pro.apex.exchange.com'.
    :type endpoint: str

    """

    # def __init__(self, endpoint=None, api_key=None, api_secret=None,
    #             logging_level=logging.INFO, log_requests=False,
    #             request_timeout=10, recv_window=5000, force_retry=False,
    #             retry_codes=None, ignore_codes=None, max_retries=3,
    #             retry_delay=3, referral_id=None, spot=False):
    #    """Initializes the HTTP class."""

    def __init__(
            self,
            endpoint,
            api_timeout=3000,  # TODO: Actually use this.
            default_ethereum_address=None,
            eth_private_key=None,
            eth_send_options=None,
            network_id=NETWORKID_MAIN,
            env_id=REGISTER_ENVID_MAIN,
            stark_private_key=None,
            stark_public_key=None,
            stark_public_key_y_coordinate=None,
            zk_seeds=None,
            zk_l2Key=None,
            web3=None,
            web3_account=None,
            web3_provider=None,
            api_key_credentials=None,
            request_timeout=10, recv_window=5000, force_retry=False,
            retry_codes=None, max_retries=3,
            retry_delay=3, referral_id=None, proxies=None
    ):
        # Remove trailing '/' if present, from host.
        if endpoint.endswith('/'):
            self.endpoint = endpoint[:-1]
        # Set the endpoint.
        if endpoint is None:
            self.endpoint = APEX_HTTP_MAIN
        else:
            self.endpoint = endpoint

        # Setup logger.

        self.logger = logging.getLogger(__name__)

        if len(logging.root.handlers) == 0:
            # no handler on root logger set -> we add handler just for this logger to not mess with custom logic from outside
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter(fmt='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                                                   datefmt='%Y-%m-%d %H:%M:%S'
                                                   )
                                 )
            handler.setLevel(logging.INFO)
            self.logger.addHandler(handler)

        #self.logger.warning(
        #    'This HTTP class is maintained for compatibility purposes. You '
        #    'should prefer importing market-specific classes, like so: '
        #    'from apexpro import HTTP; '
        #)

        self.logger.debug('Initializing HTTP session.')
        self.log_requests = False

        self.proxies = proxies
        # Set web3 keys.
        self.eth_send_options = eth_send_options or {}
        self.stark_private_key = stark_private_key
        self.api_key_credentials = api_key_credentials
        self.stark_public_key_y_coordinate = stark_public_key_y_coordinate
        self.zk_seeds = zk_seeds
        self.zk_l2Key = zk_l2Key
        self.web3 = None
        self._eth = None
        self.eth_signer = None
        self.default_address = None
        self.network_id = None

        if web3 is not None or web3_provider is not None:
            if isinstance(web3_provider, str):
                web3_provider = Web3.HTTPProvider(web3_provider)
            self.web3 = web3 or Web3(web3_provider)
            self.eth_signer = SignWithWeb3(self.web3)
            self.default_address = self.web3.eth.defaultAccount or None
            self.network_id = self.web3.net.version

        if eth_private_key is not None or web3_account is not None:
            # May override web3 or web3_provider configuration.
            key = eth_private_key or web3_account.key
            self.eth_signer = SignWithKey(key)
            self.default_address = self.eth_signer.address

        self.default_address = default_ethereum_address or self.default_address
        self.network_id = int(
            network_id or self.network_id or REGISTER_ENVID_MAIN
        )
        self.env_id = int(
            env_id or self.env_id or REGISTER_ENVID_MAIN
        )

        # Derive the public keys.
        if stark_private_key is not None:
            self.stark_public_key, self.stark_public_key_y_coordinate = (
                private_key_to_public_key_pair_hex(stark_private_key)
            )
        if (
                stark_public_key is not None and
                stark_public_key != self.stark_public_key
        ):
            self.logger.warning('STARK public/private key mismatch')
        if (
                stark_public_key_y_coordinate is not None and
                stark_public_key_y_coordinate !=
                self.stark_public_key_y_coordinate
        ):
            self.logger.warning('STARK public/private key mismatch (y)')
        else:
            self.stark_public_key = stark_public_key
            self.stark_public_key_y_coordinate = stark_public_key_y_coordinate

        self.api_key_credentials = api_key_credentials

        self.signer = SignOnboardingAction(self.eth_signer, self.network_id)
        self.starkeySigner = SignOnboardingAction(self.eth_signer, self.env_id)

        # Set timeout.
        self.timeout = request_timeout
        self.recv_window = recv_window
        self.force_retry = force_retry
        self.max_retries = max_retries
        self.retry_delay = retry_delay

        # Set whitelist of non-fatal Apexpro status codes to retry on.
        if retry_codes is None:
            self.retry_codes = {10002, 10006, 30034, 30035, 130035, 130150}
        else:
            self.retry_codes = retry_codes

        # Initialize requests session.
        self.client = requests.Session()
        #self.client.trust_env = False
        self.client.headers.update(
            {
                'User-Agent': 'apexpro-python-sdk-' + VERSION,
                'Accept': 'application/json',
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )
        self.client.proxies = self.proxies

        # Add referral ID to header.
        if referral_id:
            self.client.headers.update({'Referer': referral_id})

    def _exit(self):
        """Closes the request session."""
        self.client.close()
        self.logger.debug('HTTP session closed.')

    def _auth(self, method, params, recv_window):
        """
        Generates authentication signature per Apexpro API specifications.

        Notes
        -------------------
        Since the POST method requires a JSONified dict, we need to ensure
        the signature uses lowercase booleans instead of Python's
        capitalized booleans. This is done in the bug fix below.

        """

        api_key = self.api_key
        api_secret = self.api_secret

        if api_key is None or api_secret is None:
            raise PermissionError('Authenticated endpoints require keys.')

        # Append required parameters.
        params['api_key'] = api_key
        params['recv_window'] = recv_window
        params['timestamp'] = int(time.time() * 10 ** 3)

        # Sort dictionary alphabetically to create querystring.
        _val = '&'.join(
            [str(k) + '=' + str(v) for k, v in sorted(params.items()) if
             (k != 'sign') and (v is not None)]
        )

        # Bug fix. Replaces all capitalized booleans with lowercase.
        if method == 'POST':
            _val = _val.replace('True', 'true').replace('False', 'false')

        # Return signature.
        return str(hmac.new(
            bytes(api_secret, 'utf-8'),
            bytes(_val, 'utf-8'), digestmod='sha256'
        ).hexdigest())

    @property
    def eth(self):
        '''
        Get the eth module, used for interacting with Ethereum smart contracts.
        '''
        collateral_asset_id = ''
        token_contracts = ''
        web3_provider = ''
        starware_perpetuals_contract = self.config.get('global').get('starkExContractAddress')
        for k, v1 in enumerate(self.config.get('currency')):
            if v1.get('id') == 'USDC':
                collateral_asset_id = v1.get('starkExAssetId')
        for k, v2 in enumerate(self.config.get('multiChain').get('chains')):
            if v2.get('chainId') == self.network_id:
                web3_provider = v2.get('rpcUrl')
                for k, v3 in enumerate(v2.get('tokens')):
                    if v3.get('token') == 'USDC':
                        token_contracts = v3.get('tokenAddress')

        web3_provider = Web3.HTTPProvider(web3_provider)
        self.web3 =  Web3(web3_provider)

        if not self._eth:
            eth_private_key = getattr(self.eth_signer, '_private_key', None)
            if self.web3 and eth_private_key:
                self._eth = Eth(
                    web3=self.web3,
                    network_id=self.network_id,
                    eth_private_key=eth_private_key,
                    default_address=self.default_address,
                    stark_public_key=self.stark_public_key,
                    send_options=self.eth_send_options,
                    collateral_asset_id = collateral_asset_id,
                    starware_perpetuals_contract = starware_perpetuals_contract,
                    token_contracts = token_contracts,
                )
            else:
                raise Exception(
                    'Eth module is not supported since neither web3 ' +
                    'nor web3_provider was provided OR since neither ' +
                    'eth_private_key nor web3_account was provided',
                    )
        return self._eth

    def _verify_string(self, params, key):
        if key in params:
            if not isinstance(params[key], str):
                return False
            else:
                return True
        return True

    def configs(self, **kwargs):
        suffix = URL_SUFFIX + "/v1/symbols"
        configs = self._submit_request(
            method='GET',
            path=self.endpoint + suffix
        )
        self.env_id = configs['data']['global']['registerEnvId']
        self.config = configs['data']
        self.starkeySigner = SignOnboardingAction(self.eth_signer, self.env_id)
        return configs

    def configs_v2(self, **kwargs):
        suffix = URL_SUFFIX + "/v2/symbols"
        configs = self._submit_request(
            method='GET',
            path=self.endpoint + suffix
        )
        self.configV2 = configs['data']
        self.env_id = configs['data']['usdcConfig']['global']['registerEnvId']
        self.usdcConfigV2 = configs['data']['usdcConfig']
        self.usdtConfigV2 = configs['data']['usdtConfig']
        self.starkeySigner = SignOnboardingAction(self.eth_signer, self.env_id)
        return configs

    def configs_v3(self, **kwargs):
        suffix = URL_SUFFIX + "/v3/symbols"
        configs = self._submit_request(
            method='GET',
            path=self.endpoint + suffix
        )
        self.configV3 = configs['data']
        return configs

    def _submit_request(self, method=None, path=None, query=None, headers=None):
        """
        Submits the request to the API.

        Notes
        -------------------
        We use the params argument for the GET method, and data argument for
        the POST method. Dicts passed to the data argument must be
        JSONified prior to submitting request.

        """

        if query is None:
            query = {}

        # Store original recv_window.
        recv_window = self.recv_window

        # Send request and return headers with body. Retry if failed.
        retries_attempted = self.max_retries
        req_params = None

        while True:

            retries_attempted -= 1
            if retries_attempted < 0:
                raise FailedRequestError(
                    request=f'{method} {path}: {req_params}',
                    message='Bad Request. Retries exceeded maximum.',
                    status_code=400,
                    time=dt.utcnow().strftime("%H:%M:%S")
                )

            retries_remaining = f'{retries_attempted} retries remain.'

            # Define parameters and log the request.
            if query is not None:
                req_params = {k: v for k, v in query.items() if
                              v is not None}

            else:
                req_params = {}

            # Log the request.
            if self.log_requests:
                self.logger.debug(f'Request -> {method} {path}: {req_params}')

            # Prepare request; use 'params' for GET and 'data' for POST.
            if method == 'GET':
                r = self.client.prepare_request(
                    requests.Request(method, path, params=req_params,
                                     headers=headers)
                )
            elif method == 'POST':
                r = self.client.prepare_request(
                    requests.Request(method, path,
                                     data=req_params,
                                     headers=headers)
                )

            # Attempt the request.
            try:
                s = self.client.send(r, timeout=self.timeout)

            # If requests fires an error, retry.
            except (
                    requests.exceptions.ReadTimeout,
                    requests.exceptions.SSLError,
                    requests.exceptions.ConnectionError
            ) as e:
                if self.force_retry:
                    self.logger.error(f'{e}. {retries_remaining}')
                    time.sleep(self.retry_delay)
                    continue
                else:
                    raise e

            # Convert response to dictionary, or raise if requests error.
            try:
                s_json = s.json()

            # If we have trouble converting, handle the error and retry.
            except JSONDecodeError as e:
                if self.force_retry:
                    self.logger.error(f'{e}. {retries_remaining}')
                    time.sleep(self.retry_delay)
                    continue
                else:
                    raise FailedRequestError(
                        request=f'{method} {path}: {req_params}',
                        message='Conflict. Could not decode JSON.',
                        status_code=409,
                        time=dt.utcnow().strftime("%H:%M:%S")
                    )
            else:
                return s_json



import web3
from web3 import Web3
from web3.auto import w3
from apexpro import constants

PREPEND_DEC = '\x19Ethereum Signed Message:\n32'
PREPEND_HEX = '\x19Ethereum Signed Message:\n\x20'


def is_valid_sig_type(
    sig_type,
):
    return sig_type in [
        constants.SIGNATURE_TYPE_DECIMAL,
        constants.SIGNATURE_TYPE_HEXADECIMAL,
        constants.SIGNATURE_TYPE_NO_PREPEND,
        constants.SIGNATURE_TYPE_PERSONAL,
    ]


def ec_recover_typed_signature(
    hashVal,
    typed_signature,
):
    if len(strip_hex_prefix(typed_signature)) != 66 * 2:
        raise Exception('Unable to ecrecover signature: ' + typed_signature)

    sig_type = int(typed_signature[-2:], 16)
    prepended_hash = ''
    if sig_type == constants.SIGNATURE_TYPE_NO_PREPEND:
        prepended_hash = hashVal
    elif sig_type == constants.SIGNATURE_TYPE_DECIMAL:
        prepended_hash = Web3.solidityKeccak(
            ['string', 'bytes32'],
            [PREPEND_DEC, hashVal],
        )
    elif sig_type == constants.SIGNATURE_TYPE_HEXADECIMAL:
        prepended_hash = Web3.solidityKeccak(
            ['string', 'bytes32'],
            [PREPEND_HEX, hashVal],
        )
    else:
        raise Exception('Invalid signature type: ' + sig_type)

    if not prepended_hash:
        raise Exception('Invalid hash: ' + hashVal)

    signature = typed_signature[:-2]

    address = w3.eth.account.recoverHash(prepended_hash, signature=signature)
    return address


def create_typed_signature(
    signature,
    sig_type,
):
    if not is_valid_sig_type(sig_type):
        raise Exception('Invalid signature type: ' + sig_type)

    return fix_raw_signature(signature) + '0' + str(sig_type)


def fix_raw_signature(
    signature,
):
    stripped = strip_hex_prefix(signature)

    if len(stripped) != 130:
        raise Exception('Invalid raw signature: ' + signature)

    rs = stripped[:128]
    v = stripped[128: 130]

    if v == '00':
        return '0x' + rs + '1b'
    if v == '01':
        return '0x' + rs + '1c'
    if v in ['1b', '1c']:
        return '0x' + stripped

    raise Exception('Invalid v value: ' + v)

# ============ Byte Helpers ============


def strip_hex_prefix(input):
    if input.startswith('0x'):
        return input[2:]

    return input


def addresses_are_equal(
    address_one,
    address_two,
):
    if not address_one or not address_two:
        return False

    return (
        strip_hex_prefix(
            address_one
        ).lower() == strip_hex_prefix(address_two).lower()
    )


def hash_string(input):
    return Web3.solidityKeccak(['string'], [input])

def hash_person(input):
    msg = "\x19Ethereum Signed Message:\n%d%s" %( len(input), input)
    return Web3.solidityKeccak(['string'], [msg])



from web3 import Web3

from apexpro.constants import REGISTER_ENVID_MAIN
from apexpro.eth_signing import util
from apexpro.eth_signing.sign_off_chain_action import SignOffChainAction

# On mainnet, include an extra onlySignOn parameter.
EIP712_ONBOARDING_ACTION_STRUCT = [
    {'type': 'string', 'name': 'action'},
    {'type': 'string', 'name': 'onlySignOn'},
    {'type': 'string', 'name': 'nonce'},
]
EIP712_ONBOARDING_ACTION_STRUCT_STRING = (
    'ApeX(' +
    'string action,' +
    'string onlySignOn,' +
    'string nonce' +
    ')'
)
EIP712_ONBOARDING_ACTION_STRUCT_TESTNET = [
    {'type': 'string', 'name': 'action'},
]
EIP712_ONBOARDING_ACTION_STRUCT_STRING_TESTNET = (
    'apex(' +
    'string action' +
    ')'
)
EIP712_STRUCT_NAME = 'ApeX'

ONLY_SIGN_ON_DOMAIN_MAINNET = 'https://pro.apex.exchange'


class SignOnboardingAction(SignOffChainAction):

    def get_eip712_struct(self):
        # On mainnet, include an extra onlySignOn parameter.
        #if self.network_id == REGISTER_ENVID_MAIN:
        return EIP712_ONBOARDING_ACTION_STRUCT
        #else:
        #    return EIP712_ONBOARDING_ACTION_STRUCT_TESTNET

    def get_eip712_struct_name(self):
        return EIP712_STRUCT_NAME

    def get_eip712_message(
        self,
        **message,
    ):
        eip712_message = super(SignOnboardingAction, self).get_eip712_message(
            **message,
        )

        # On mainnet, include an extra onlySignOn parameter.
        #if self.network_id == REGISTER_ENVID_MAIN:
        eip712_message['message']['onlySignOn'] = (
                'https://pro.apex.exchange'
        )

        return eip712_message

    def get_hash(
        self,
        action,
        nonce
    ):
        # On mainnet, include an extra onlySignOn parameter.
        #if self.network_id == REGISTER_ENVID_MAIN:
        eip712_struct_str = EIP712_ONBOARDING_ACTION_STRUCT_STRING
        #else:
        #    eip712_struct_str = EIP712_ONBOARDING_ACTION_STRUCT_STRING_TESTNET

        data = [
            [
                'bytes32',
                'bytes32',
            ],
            [
                util.hash_string(eip712_struct_str),
                util.hash_string(action)
            ],
        ]

        # On mainnet, include an extra onlySignOn parameter.
        #if self.network_id == REGISTER_ENVID_MAIN:
        data[0].append('bytes32')
        data[1].append(util.hash_string(ONLY_SIGN_ON_DOMAIN_MAINNET))

        data[0].append('bytes32')
        data[1].append(util.hash_string(nonce))

        struct_hash = Web3.solidityKeccak(*data)
        return self.get_eip712_hash(struct_hash)



import base64
import hashlib

import websocket
import threading
import time
import json
import hmac
import logging
import re
import copy
from . import HTTP, APEX_HTTP_MAIN
from .constants import APEX_WS_MAIN
from .helpers.request_helpers import generate_now


logger = logging.getLogger(__name__)

PRIVATE_REQUEST_PATH = '/ws/accounts'
PRIVATE_WSS = "/realtime_private?v=2"
PUBLIC_WSS = "/realtime_public?v=2"

class _WebSocketManager:
    def __init__(self, callback_function, endpoint="", api_key_credentials=None,
                 ping_interval=15, ping_timeout=None,
                 restart_on_error=True, trace_logging=False):

        if endpoint.endswith('/'):
            self.endpoint = endpoint[:-1]
        # Set the endpoint.
        if endpoint is None:
            self.endpoint = APEX_WS_MAIN
        else:
            self.endpoint = endpoint

        self.api_key_credentials = api_key_credentials

        self.callback = callback_function

        # Setup the callback directory following the format:
        #   {
        #       "topic_name": function
        #   }
        self.callback_directory = {}

        # Set ping settings.
        self.ping_interval = ping_interval
        self.ping_timeout = ping_timeout

        # Other optional data handling settings.
        self.handle_error = restart_on_error
        self.timer = None
        self.r_timer = None

        # Enable websocket-client's trace logging for extra debug information
        # on the websocket connection, including the raw sent & recv messages
        websocket.enableTrace(trace_logging)

        # Set initial state, initialize dictionary and connect.
        self._reset()

    def sign(
            self,
            request_path,
            method,
            iso_timestamp,
    ):

        message_string = (
                iso_timestamp +
                method +
                request_path
        )
        hashed = hmac.new(
            base64.standard_b64encode(
                (self.api_key_credentials['secret']).encode(encoding='utf-8'),
            ),
            msg=message_string.encode(encoding='utf-8'),
            digestmod=hashlib.sha256,
        )
        return base64.standard_b64encode(hashed.digest()).decode()

    def runTimer(self):
        time_stamp = generate_now()
        ping = json.dumps({
            "op": "ping",
            "args": [str(time_stamp)]
        })
        self.ws.send(ping)
        #print("send ping:" + ping)

    def _on_open(self):
        """
        Log WS open.
        """

        logger.debug(f"WebSocket  opened.")

    def _on_message(self, message):
        """
        Parse incoming messages.
        """
        self.callback(json.loads(message))

    def _connect(self, url):
        """
        Open websocket in a thread.
        """
        self.private_websocket = True if url.__contains__("private") else False

        time_stamp = generate_now()
        self.ws = websocket.WebSocketApp(
            url=url + '&timestamp=' + str(time_stamp),
            on_message=lambda ws, msg: self._on_message(msg),
            on_close=self._on_close(),
            on_open=self._on_open(),
            on_error=lambda ws, err: self._on_error(err)
        )

        # Setup the thread running WebSocketApp.
        self.wst = threading.Thread(target=lambda: self.ws.run_forever(
            ping_interval=self.ping_interval,
            ping_timeout=self.ping_timeout
        ))

        # Configure as daemon; start.
        self.wst.daemon = True
        self.wst.start()

        # Attempt to connect for X seconds.
        retries = 10
        while retries > 0 and (not self.ws.sock or not self.ws.sock.connected):
            retries -= 1
            time.sleep(1)

        # If connection was not successful, raise error.
        if retries <= 0:
            self.exit()
            raise websocket.WebSocketTimeoutException("Connection failed.")

        # If given an api_key, authenticate.
        if self.api_key_credentials:
            self._auth(time_stamp)

    def _auth(self, time_stamp):
        """
        Authorize websocket connection.
        """

        signature = self.sign(
            request_path=PRIVATE_REQUEST_PATH,
            method='GET',
            iso_timestamp=str(time_stamp),
        )


        req = {
            'type': 'login',
            'topics': ['ws_notify_v1', 'ws_accounts_v2', 'ws_zk_accounts_v3','ws_accounts_v1'],
            'httpMethod': 'GET',
            'requestPath': PRIVATE_REQUEST_PATH,
            'apiKey': self.api_key_credentials['key'],
            'passphrase': self.api_key_credentials['passphrase'],
            'timestamp': time_stamp,
            'signature': signature,
        }
        sendStr = \
            {
                "op": "login",
                "args": [json.dumps(req)]
            }

        # Authenticate with API.
        self.ws.send(
            json.dumps(sendStr)
        )

    def _on_error(self, error):
        """
        Exit on errors and raise exception, or attempt reconnect.
        """

        if not self.exited:
            logger.error(f"WebSocket  encountered error: {error}.")
            self.exit()

        # Reconnect.
        if self.handle_error:
            self._reset()
            if self.private_websocket:
                self._connect(self.endpoint + PRIVATE_WSS)
            else:
                self._connect(self.endpoint + PUBLIC_WSS)

    def _on_close(self):
        """
        Log WS close.
        """
        logger.debug(f"WebSocket closed.")

    def _reset(self):
        """
        Set state booleans and initialize dictionary.
        """
        self.exited = False
        self.auth = False
        self.data = {}

    def exit(self):
        """
        Closes the websocket connection.
        """

        self.ws.close()
        while self.ws.sock:
            continue
        self.exited = True


class _ApexWebSocketManager(_WebSocketManager):
    def __init__(self, **kwargs):
        super().__init__(self._handle_incoming_message, **kwargs)

    def subscribe(self, sendStr, topic, callback):
        """
        Formats and sends the subscription message, given a topic. Saves the
        provided callback function, to be called by incoming messages.
        """

        if self.private_websocket:
            # Spot private topics don't need a subscription message
            self._set_callback(topic, callback)
            return

        self.ws.send(sendStr)
        self._set_callback(topic, callback)

    def _handle_incoming_message(self, message):
        #print(message)
        def is_ping_message():
            if type(message) == dict and message.get("op") == "ping":
                return True
            else:
                return False

        def is_auth_message():
            if type(message) == dict and \
                    message.get("request") is not None and message.get("request").get("op") is not None and message.get("request").get("op") == "login":
                return True
            else:
                return False

        def is_subscription_message():
            if type(message) == dict and \
                    message.get("request") is not None and message.get("request").get("op") is not None and message.get("request").get("op") == "subscribe":
                return True
            else:
                return False

        if is_ping_message():
            time_stamp = generate_now()
            pong = json.dumps({
                "op": "pong",
                "args": [str(time_stamp)]
            })
            self.ws.send(pong)
            #print("send pong:" + pong)
            return

        # Check auth
        if is_auth_message():
            # If we get successful spot auth, notify user
            if message.get("success") is not None and message.get("success") == "true":
                logger.debug(f"Authorization successful.")
                self.auth = True

        # Check subscription
        elif is_subscription_message():
            # If we get successful spot subscription, notify user
            if message.get("success") is not None and message.get("success") == "true":
                logger.debug(f"Subscription successful.")


        else:  # Standard topic push
                if message.get('topic') is not None and self.callback_directory.get(message.get('topic')) is not None:
                    callback_function = self.callback_directory[message['topic'] ]
                    callback_function(message)

    def _set_callback(self, topic, callback_function):
        self.callback_directory[topic] = callback_function

    def _get_callback(self, topic):
        return self.callback_directory[topic]

    def _pop_callback(self, topic):
        self.callback_directory.pop(topic)

    def _check_callback_directory(self, topics):
        for topic in topics:
            if topic in self.callback_directory:
                raise Exception(f"You have already subscribed to this topic: "
                                f"{topic}")


def _identify_ws_method(input_wss_url, wss_dictionary):
    """
    This method matches the input_wss_url with a particular WSS method. This
    helps ensure that, when subscribing to a custom topic, the topic
    subscription message is sent down the correct WSS connection.
    """
    path = re.compile("(wss://)?([^/\s]+)(.*)")
    input_wss_url_path = path.match(input_wss_url).group(3)
    for wss_url, function_call in wss_dictionary.items():
        wss_url_path = path.match(wss_url).group(3)
        if input_wss_url_path == wss_url_path:
            return function_call


def _find_index(source, target, key):
    """
    Find the index in source list of the targeted ID.
    """
    return next(i for i, j in enumerate(source) if j[key] == target[key])


def _make_public_kwargs(private_kwargs):
    public_kwargs = copy.deepcopy(private_kwargs)
    public_kwargs.pop("api_key_credentials", "")
    return public_kwargs












import eth_account

from apexpro.constants import SIGNATURE_TYPE_NO_PREPEND, SIGNATURE_TYPE_PERSONAL
from apexpro.eth_signing import util


class Signer(object):

    def sign(
        self,
        eip712_message,
        message_hash,
        opt_signer_address,
    ):
        '''
        Sign an EIP-712 message.

        Returns a “typed signature” whose last byte indicates whether the hash
        was prepended before being signed.

        :param eip712_message: required
        :type eip712_message: dict

        :param message_hash: required
        :type message_hash: HexBytes

        :param opt_signer_address: optional
        :type opt_signer_address: str

        :returns: str
        '''
        raise NotImplementedError()


class SignWithWeb3(Signer):

    def __init__(self, web3):
        self.web3 = web3

    def sign(
        self,
        eip712_message,
        message_hash,  # Ignored.
        opt_signer_address,
    ):
        signer_address = opt_signer_address or self.web3.eth.defaultAccount
        if not signer_address:
            raise ValueError(
                'Must set ethereum_address or web3.eth.defaultAccount',
            )
        raw_signature = self.web3.eth.signTypedData(
            signer_address,
            eip712_message,
        )
        typed_signature = util.create_typed_signature(
            raw_signature.hex(),
            SIGNATURE_TYPE_NO_PREPEND,
        )
        return typed_signature


class SignWithKey(Signer):

    def __init__(self, private_key):
        self.address = eth_account.Account.from_key(private_key).address
        self._private_key = private_key

    def sign(
        self,
        eip712_message,  # Ignored.
        message_hash,
        opt_signer_address,
    ):
        if (
            opt_signer_address is not None and
            opt_signer_address != self.address
        ):
            raise ValueError(
                'signer_address is {} but Ethereum key (eth_private_key / '
                'web3_account) corresponds to address {}'.format(
                    opt_signer_address,
                    self.address,
                ),
            )
        signed_message = eth_account.Account._sign_hash(
            message_hash.hex(),
            self._private_key,
        )
        typed_signature = util.create_typed_signature(
            signed_message.signature.hex(),
            SIGNATURE_TYPE_NO_PREPEND,
        )
        return typed_signature

    def sign_person(
            self,
            eip712_message,  # Ignored.
            message_hash,
            opt_signer_address,
    ):
        if (
                opt_signer_address is not None and
                opt_signer_address != self.address
        ):
            raise ValueError(
                'signer_address is {} but Ethereum key (eth_private_key / '
                'web3_account) corresponds to address {}'.format(
                    opt_signer_address,
                    self.address,
                ),
            )

        signed_message = eth_account.Account._sign_hash(
            message_hash.hex(),
            self._private_key,
        )
        typed_signature = util.create_typed_signature(
            signed_message.signature.hex(),
            SIGNATURE_TYPE_PERSONAL,
        )
        return typed_signature

    def sign_zk_person(
            self,
            message_hash,
            opt_signer_address,
    ):
        if (
                opt_signer_address is not None and
                opt_signer_address != self.address
        ):
            raise ValueError(
                'signer_address is {} but Ethereum key (eth_private_key / '
                'web3_account) corresponds to address {}'.format(
                    opt_signer_address,
                    self.address,
                ),
            )

        signed_message = eth_account.Account._sign_hash(
            message_hash.hex(),
            self._private_key,
        )
        signature = util.fix_raw_signature(
            signed_message.signature.hex()
        )
        return signature



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='apexomni-x86-windows-linux',
    version='3.0.2',
    packages=find_packages(),
    description='Python3 Apexpro omni HTTP/WebSocket API Connector',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ApeX-Protocol/apexpro-openapi',
    license='MIT License',
    author='Dexter Dickinson',
    author_email='xxx@apexomni.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='apexomni apexpro api connector',
    package_data={'': ['*.json','*.dylib','*.dll','*.so']},
    data_files=[('apexpro', ['apexpro/starkex/starkex_resources/pedersen_params.json', 'apexpro/libzklink_sdk.dylib'])],
    # packages=['apexpro'],
    python_requires='>=3.6',
    install_requires=[
        'requests',
        'websocket-client',
        'websockets',
        'dateparser==1.0.0',
        'ecdsa==0.16.0',
        'eth_keys',
        'eth-account>=0.4.0,<0.6.0',
        'mpmath==1.0.0',
        'pytest>=4.4.0,<5.0.0',
        'requests-mock==1.6.0',
        'requests>=2.22.0,<3.0.0',
        'setuptools>=50.3.2,<70.6.0',
        'sympy==1.6',
        'tox==3.13.2',
        'web3>=5.0.0,<6.0.0'
    ],
)



import json
import os
from web3 import Web3

from apexpro.constants import ASSET_RESOLUTION
from apexpro.constants import COLLATERAL_ASSET
from apexpro.constants import DEFAULT_GAS_AMOUNT
from apexpro.constants import DEFAULT_GAS_MULTIPLIER
from apexpro.constants import DEFAULT_GAS_PRICE
from apexpro.constants import DEFAULT_GAS_PRICE_ADDITION
from apexpro.constants import MAX_SOLIDITY_UINT
from apexpro.errors import TransactionReverted

ERC20_ABI = 'abi/erc20.json'
STARKWARE_PERPETUALS_ABI = 'abi/starkware-perpetuals.json'
COLLATERAL_ASSET_RESOLUTION = float(ASSET_RESOLUTION[COLLATERAL_ASSET])


class Eth(object):

    def __init__(
        self,
        web3,
        network_id,
        eth_private_key,
        default_address,
        stark_public_key,
        send_options,
        collateral_asset_id,
        starware_perpetuals_contract,
        token_contracts,
    ):
        self.web3 = web3
        self.network_id = network_id
        self.eth_private_key = eth_private_key
        self.default_address = default_address
        self.stark_public_key = stark_public_key
        self.send_options = send_options

        self.collateral_asset_id = collateral_asset_id
        self.starware_perpetuals_contract = starware_perpetuals_contract
        self.token_contracts = token_contracts

        self.cached_contracts = {}
        self._next_nonce_for_address = {}

    # -----------------------------------------------------------
    # Helper Functions
    # -----------------------------------------------------------

    def create_contract(
        self,
        address,
        file_path,
    ):

        apex_folder = os.path.dirname(os.path.abspath(__file__))
        return self.web3.eth.contract(
            address=address,
            abi=json.load(open(os.path.join(apex_folder, file_path), 'r')),
        )

    def get_contract(
        self,
        address,
        file_path,
    ):
        if address not in self.cached_contracts:
            self.cached_contracts[address] = self.create_contract(
                address,
                file_path,
            )
        return self.cached_contracts[address]

    def get_exchange_contract(
        self,
        contract_address=None,
    ):
        if contract_address is None:
            contract_address = self.starware_perpetuals_contract
        if contract_address is None:
            raise ValueError(
                'Perpetuals exchange contract on network {}'.format(
                    self.network_id,
                )
            )
        contract_address = Web3.toChecksumAddress(contract_address)
        return self.get_contract(contract_address, STARKWARE_PERPETUALS_ABI)

    def get_token_contract(
        self,
        asset,
        token_address,
    ):
        if token_address is None:
            token_address = self.token_contracts
        if token_address is None:
            raise ValueError(
                'Token address unknown for asset {} on network {}'.format(
                    asset,
                    self.network_id,
                )
            )
        token_address = Web3.toChecksumAddress(token_address)
        return self.get_contract(token_address, ERC20_ABI)

    def send_eth_transaction(
        self,
        method=None,
        options=None,
    ):
        options = dict(self.send_options, **(options or {}))

        if 'from' not in options:
            options['from'] = self.default_address
        if options.get('from') is None:
            raise ValueError(
                "options['from'] is not set, and no default address is set",
            )
        auto_detect_nonce = 'nonce' not in options
        if auto_detect_nonce:
            options['nonce'] = self.get_next_nonce(options['from'])
        if 'gasPrice' not in options:
            try:
                options['gasPrice'] = (
                    self.web3.eth.gasPrice + DEFAULT_GAS_PRICE_ADDITION
                )
            except Exception:
                options['gasPrice'] = DEFAULT_GAS_PRICE
        if 'value' not in options:
            options['value'] = 0
        gas_multiplier = options.pop('gasMultiplier', DEFAULT_GAS_MULTIPLIER)
        if 'gas' not in options:
            try:
                options['gas'] = int(
                    method.estimateGas(options) * gas_multiplier
                )
            except Exception:
                options['gas'] = DEFAULT_GAS_AMOUNT

        signed = self.sign_tx(method, options)
        try:
            tx_hash = self.web3.eth.sendRawTransaction(signed.rawTransaction)
        except ValueError as error:
            while (
                auto_detect_nonce and
                (
                    'nonce too low' in str(error) or
                    'replacement transaction underpriced' in str(error)
                )
            ):
                try:
                    options['nonce'] += 1
                    signed = self.sign_tx(method, options)
                    tx_hash = self.web3.eth.sendRawTransaction(
                        signed.rawTransaction,
                    )
                except ValueError as inner_error:
                    error = inner_error
                else:
                    break  # Break on success...
            else:
                raise error  # ...and raise error otherwise.

        # Update next nonce for the account.
        self._next_nonce_for_address[options['from']] = options['nonce'] + 1

        return tx_hash.hex()

    def get_next_nonce(
        self,
        ethereum_address,
    ):
        if self._next_nonce_for_address.get(ethereum_address) is None:
            self._next_nonce_for_address[ethereum_address] = (
                self.web3.eth.getTransactionCount(ethereum_address)
            )
        return self._next_nonce_for_address[ethereum_address]

    def sign_tx(
        self,
        method,
        options,
    ):
        if method is None:
            tx = options
        else:
            tx = method.buildTransaction(options)
        return self.web3.eth.account.sign_transaction(
            tx,
            self.eth_private_key,
        )

    def wait_for_tx(
        self,
        tx_hash,
    ):
        '''
        Wait for a tx to be mined and return the receipt. Raise on revert.

        :param tx_hash: required
        :type tx_hash: number

        :returns: transactionReceipt

        :raises: TransactionReverted
        '''
        tx_receipt = self.web3.eth.waitForTransactionReceipt(tx_hash)
        if tx_receipt['status'] == 0:
            raise TransactionReverted(tx_receipt)

    # -----------------------------------------------------------
    # Transactions
    # -----------------------------------------------------------

    def register_user(
        self,
        registration_signature,
        stark_public_key=None,
        ethereum_address=None,
        send_options=None,
    ):
        '''
        Register a STARK key, using a signature provided by dYdX.

        :param registration_signature: required
        :type registration_signature: string

        :param stark_public_key: optional
        :type stark_public_key: string

        :param ethereum_address: optional
        :type ethereum_address: string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        stark_public_key = stark_public_key or self.stark_public_key
        if stark_public_key is None:
            raise ValueError('No stark_public_key was provided')

        ethereum_address = ethereum_address or self.default_address
        if ethereum_address is None:
            raise ValueError(
                'ethereum_address was not provided, '
                'and no default address is set',
            )

        contract = self.get_exchange_contract()
        return self.send_eth_transaction(
            method=contract.functions.registerUser(
                ethereum_address,
                int(stark_public_key, 16),
                registration_signature,
            ),
            options=send_options,
        )

    def deposit_to_exchange(
        self,
        position_id,
        human_amount,
        stark_public_key=None,
        send_options=None,
    ):
        '''
        Deposit collateral to the L2 perpetuals exchange.

        :param position_id: required
        :type position_id: int or string

        :param human_amount: required
        :type human_amount: number or string

        :param stark_public_key: optional
        :type stark_public_key: string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        stark_public_key = stark_public_key or self.stark_public_key
        if stark_public_key is None:
            raise ValueError('No stark_public_key was provided')

        contract = self.get_exchange_contract()
        return self.send_eth_transaction(
            method=contract.functions.deposit(
                int(stark_public_key, 16),
                int(self.collateral_asset_id, 16),
                int(position_id),
                int(float(human_amount) * COLLATERAL_ASSET_RESOLUTION),
            ),
            options=send_options,
        )

    def withdraw(
        self,
        stark_public_key=None,
        send_options=None,
    ):
        '''
        Withdraw from exchange.

        :param stark_public_key: optional
        :type stark_public_key: string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        stark_public_key = stark_public_key or self.stark_public_key
        if stark_public_key is None:
            raise ValueError('No stark_public_key was provided')

        contract = self.get_exchange_contract()
        return self.send_eth_transaction(
            method=contract.functions.withdraw(
                int(stark_public_key, 16),
                int(self.collateral_asset_id, 16),
            ),
            options=send_options,
        )

    def withdraw_to(
        self,
        recipient,
        stark_public_key=None,
        send_options=None,
    ):
        '''
        Withdraw from exchange to address.

        :param recipient: required
        :type recipient: string

        :param stark_public_key: optional
        :type stark_public_key: string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        stark_public_key = stark_public_key or self.stark_public_key
        if stark_public_key is None:
            raise ValueError('No stark_public_key was provided')

        contract = self.get_exchange_contract()
        return self.send_eth_transaction(
            method=contract.functions.withdrawTo(
                int(stark_public_key, 16),
                int(self.collateral_asset_id, 16),
                recipient,
            ),
            options=send_options,
        )

    def transfer_eth(
        self,
        to_address=None,  # Require keyword args to avoid confusing the amount.
        human_amount=None,
        send_options=None,
    ):
        '''
        Send Ethereum.

        :param to_address: required
        :type to_address: number

        :param human_amount: required
        :type human_amount: number or string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        if to_address is None:
            raise ValueError('to_address is required')

        if human_amount is None:
            raise ValueError('human_amount is required')

        return self.send_eth_transaction(
            options=dict(
                send_options,
                to=to_address,
                value=Web3.toWei(human_amount, 'ether'),
            ),
        )

    def transfer_token(
        self,
        to_address=None,  # Require keyword args to avoid confusing the amount.
        human_amount=None,
        asset=COLLATERAL_ASSET,
        token_address=None,
        send_options=None,
    ):
        '''
        Send Ethereum.

        :param to_address: required
        :type to_address: number

        :param human_amount: required
        :type human_amount: number of string

        :param asset: optional
        :type asset: string

        :param token_address: optional
        :type asset: string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        if to_address is None:
            raise ValueError('to_address is required')

        if human_amount is None:
            raise ValueError('human_amount is required')

        if asset not in ASSET_RESOLUTION:
            raise ValueError('Unknown asset {}'.format(asset))
        asset_resolution = ASSET_RESOLUTION[asset]

        contract = self.get_token_contract(asset, token_address)
        return self.send_eth_transaction(
            method=contract.functions.transfer(
                to_address,
                int(float(human_amount) * float(asset_resolution)),
            ),
            options=send_options,
        )

    def set_token_max_allowance(
        self,
        spender,
        asset=COLLATERAL_ASSET,
        token_address=None,
        send_options=None,
    ):
        '''
        Set max allowance for some spender for some asset or token_address.

        :param spender: required
        :type spender: string

        :param asset: optional
        :type asset: string

        :param token_address: optional
        :type asset: string

        :param send_options: optional
        :type send_options: dict

        :returns: transactionHash

        :raises: ValueError
        '''
        contract = self.get_token_contract(asset, token_address)
        return self.send_eth_transaction(
            method=contract.functions.approve(
                spender,
                MAX_SOLIDITY_UINT,
            ),
            options=send_options,
        )

    # -----------------------------------------------------------
    # Getters
    # -----------------------------------------------------------

    def get_eth_balance(
        self,
        owner=None,
    ):
        '''
        Get the owner's ether balance as a human readable amount.

        :param owner: optional
        :type owner: string

        :returns: string

        :raises: ValueError
        '''
        owner = owner or self.default_address
        if owner is None:
            raise ValueError(
                'owner was not provided, and no default address is set',
            )

        wei_balance = self.web3.eth.getBalance(owner)
        return Web3.fromWei(wei_balance, 'ether')

    def get_token_balance(
        self,
        owner=None,
        asset=COLLATERAL_ASSET,
        token_address=None,
    ):
        '''
        Get the owner's balance for some asset or token address.

        :param owner: optional
        :type owner: string

        :param asset: optional
        :type asset: string

        :param token_address: optional
        :type asset: string

        :returns: int
        '''
        owner = owner or self.default_address
        if owner is None:
            raise ValueError(
                'owner was not provided, and no default address is set',
            )

        contract = self.get_token_contract(asset, token_address)
        return contract.functions.balanceOf(owner).call()

    def get_token_allowance(
        self,
        spender,
        owner=None,
        asset=COLLATERAL_ASSET,
        token_address=None,
    ):
        '''
        Get allowance for some spender for some asset or token address.

        :param spender: required
        :type spender: string

        :param owner: optional
        :type owner: string

        :param asset: optional
        :type asset: string

        :param token_address: optional
        :type token_address: string

        :returns: int

        :raises: ValueError
        '''
        owner = owner or self.default_address
        if owner is None:
            raise ValueError(
                'owner was not provided, and no default address is set',
            )

        contract = self.get_token_contract(asset, token_address)
        return contract.functions.allowance(owner, spender).call()



import decimal
import hashlib
import time

import numpy as np

import zklink_sdk as sdk
from apexpro.constants import URL_SUFFIX, ORDER_SIDE_BUY
from apexpro.helpers.request_helpers import random_client_id
from apexpro.http_private_v3 import HttpPrivate_v3
from apexpro.starkex.order import DECIMAL_CONTEXT_ROUND_UP, DECIMAL_CONTEXT_ROUND_DOWN


class HttpPrivateSign(HttpPrivate_v3):
    def create_order_v3(self,
                     symbol,
                     side,
                     type,
                     size,
                     subAccountId=None,
                     takerFeeRate=None,
                     makerFeeRate=None,
                     price=None,
                     accountId=None,
                     timeInForce="GOOD_TIL_CANCEL",
                     reduceOnly=False,
                     triggerPrice=None,
                     triggerPriceType=None,
                     trailingPercent=None,
                     clientId=None,
                     timestampSeconds=None,
                     isPositionTpsl = False,
                     signature=None,
                     isOpenTpslOrder=False,
                     isSetOpenSl=False,
                     isSetOpenTp=False,
                     slClientId=None,
                     slPrice=None,
                     slSide=None,
                     slSize=None,
                     slTriggerPrice=None,
                     tpClientId=None,
                     tpPrice=None,
                     tpSide=None,
                     tpSize=None,
                     tpTriggerPrice=None,
                     sourceFlag=None,
                     brokerId=None,):
        """"
        POST  create_order.
        client.create_order(symbol="BTC-USDT", side="SELL",
                                           type="LIMIT", size="0.01",
                                           price="20000")

       :returns: Request results as dictionary.
       """
        price = str(price)
        size = str(size)
        clientId = clientId or random_client_id()

        accountId = accountId or self.accountV3.get('id')
        if not accountId:
            raise Exception(
                'No accountId provided' +
                'please call get_account_v3()'
            )

        if not self.configV3:
            raise Exception(
                'No config provided' +
                'please call configs_v3()'
            )
        symbolData = None
        currency = {}
        for k, v in enumerate(self.configV3.get('contractConfig').get('perpetualContract')):
            if v.get('symbol') == symbol or v.get('symbolDisplayName') == symbol:
                symbolData = v
        if symbolData is None:
            for k, v in enumerate(self.configV3.get('contractConfig').get('prelaunchContract')):
                if v.get('symbol') == symbol or v.get('symbolDisplayName') == symbol:
                    symbolData = v

        for k, v2 in enumerate(self.configV3.get('contractConfig').get('assets')):
            if v2.get('token') == symbolData.get('settleAssetId'):
                currency = v2

        if symbolData is not None :
            number = decimal.Decimal(price) / decimal.Decimal(symbolData.get('tickSize'))
            if number > int(number):
                raise Exception(
                    'the price must Multiple of tickSize'
                )

        if not self.zk_seeds:
            raise Exception(
                'No signature provided and client was not ' +
                'initialized with zk_seeds'
            )

        timestampSeconds = timestampSeconds or int(time.time())
        timestampSeconds = int(timestampSeconds + 3600 * 24 * 28)


        subAccountId = subAccountId or self.accountV3.get('spotAccount').get('defaultSubAccountId')
        takerFeeRate = takerFeeRate or self.accountV3.get('contractAccount').get('takerFeeRate')
        makerFeeRate = makerFeeRate or self.accountV3.get('contractAccount').get('makerFeeRate')

        message = hashlib.sha256()
        message.update(clientId.encode())  # Encode as UTF-8.
        nonceHash = message.hexdigest()
        nonceInt = int(nonceHash, 16)

        maxUint32 = np.iinfo(np.uint32).max
        maxUint64 = np.iinfo(np.uint64).max

        slotId = (nonceInt % maxUint64)/maxUint32
        nonce = nonceInt % maxUint32
        accountId = int(accountId, 10) % maxUint32


        priceStr = (decimal.Decimal(price) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_DOWN)
        sizeStr = (decimal.Decimal(size) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_DOWN)

        takerFeeRateStr =  (decimal.Decimal(takerFeeRate) * decimal.Decimal(10000)).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)
        makerFeeRateStr =  (decimal.Decimal(makerFeeRate) * decimal.Decimal(10000)).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)

        builder = sdk.ContractBuilder(
            int(accountId),  int(subAccountId), int(slotId), int(nonce),  int(symbolData.get('l2PairId')), sizeStr.__str__(), priceStr.__str__(), side == "BUY",  int(takerFeeRateStr), int(makerFeeRateStr),  False
        )


        tx = sdk.Contract(builder)
        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)
        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature


        if side == ORDER_SIDE_BUY:
            human_cost = DECIMAL_CONTEXT_ROUND_UP.multiply(
                decimal.Decimal(size),
                decimal.Decimal(price)
            )
            fee = DECIMAL_CONTEXT_ROUND_UP.multiply(human_cost, decimal.Decimal(takerFeeRate))
        else:
            human_cost = DECIMAL_CONTEXT_ROUND_DOWN.multiply(
                decimal.Decimal(size),
                decimal.Decimal(price)
            )
            fee = DECIMAL_CONTEXT_ROUND_DOWN.multiply(human_cost, decimal.Decimal(takerFeeRate))

        limit_fee_rounded = DECIMAL_CONTEXT_ROUND_UP.quantize(
            decimal.Decimal(fee),
            decimal.Decimal(currency.get('showStep')), )

        sl_limit_fee_rounded = None
        slSignature = None
        slTriggerPriceType = None
        slExpiration = None
        tp_limit_fee_rounded = None
        tpSignature = None
        tpTriggerPriceType = None
        tpExpiration = None

        if isOpenTpslOrder == True:
            if isSetOpenSl == True:
                slTriggerPriceType = triggerPriceType
                slExpiration = timestampSeconds
                slClientId = slClientId or random_client_id()

                slMessage = hashlib.sha256()
                slMessage.update(slClientId.encode())  # Encode as UTF-8.
                slNonceHash = slMessage.hexdigest()
                slNonceInt = int(slNonceHash, 16)

                slSlotId = (slNonceInt % maxUint64)/maxUint32
                slNonce = slNonceInt % maxUint32

                slPriceStr = (decimal.Decimal(slPrice) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_DOWN)
                slSizeStr = (decimal.Decimal(slSize) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_DOWN)

                slBuilder = sdk.ContractBuilder(
                    int(accountId),  int(subAccountId), int(slSlotId), int(slNonce),  int(symbolData.get('l2PairId')), slSizeStr.__str__(), slPriceStr.__str__(), slSide == "BUY",  int(takerFeeRateStr), int(makerFeeRateStr),  False
                )

                slTx = sdk.Contract(slBuilder)
                sl_auth_data = signerSeed.sign_musig(slTx.get_bytes())
                slSignature = sl_auth_data.signature

                if slSide == ORDER_SIDE_BUY:
                    slHuman_cost = DECIMAL_CONTEXT_ROUND_UP.multiply(
                    decimal.Decimal(slSize),
                    decimal.Decimal(slPrice)
                    )
                    slFee = DECIMAL_CONTEXT_ROUND_UP.multiply(slHuman_cost, decimal.Decimal(takerFeeRate))
                else:
                    slHuman_cost = DECIMAL_CONTEXT_ROUND_DOWN.multiply(
                        decimal.Decimal(slSize),
                        decimal.Decimal(slPrice)
                    )
                    slFee = DECIMAL_CONTEXT_ROUND_DOWN.multiply(slHuman_cost, decimal.Decimal(takerFeeRate))

                sl_limit_fee_rounded = DECIMAL_CONTEXT_ROUND_UP.quantize(
                    decimal.Decimal(slFee),
                    decimal.Decimal(currency.get('showStep')), )

            if isSetOpenTp == True:
                tpTriggerPriceType = triggerPriceType
                tpExpiration = timestampSeconds
                tpClientId = tpClientId or random_client_id()

                tpMessage = hashlib.sha256()
                tpMessage.update(tpClientId.encode())  # Encode as UTF-8.
                tpNonceHash = tpMessage.hexdigest()
                tpNonceInt = int(tpNonceHash, 16)

                tpSlotId = (tpNonceInt % maxUint64)/maxUint32
                tpNonce = tpNonceInt % maxUint32

                tpPriceStr = (decimal.Decimal(tpPrice) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_DOWN)
                tpSizeStr = (decimal.Decimal(tpSize) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_DOWN)

                tpBuilder = sdk.ContractBuilder(
                    int(accountId),  int(subAccountId), int(tpSlotId), int(tpNonce),  int(symbolData.get('l2PairId')), tpSizeStr.__str__(), tpPriceStr.__str__(), tpSide == "BUY",  int(takerFeeRateStr), int(makerFeeRateStr),  False
                )

                tpTx = sdk.Contract(tpBuilder)
                tp_auth_data = signerSeed.sign_musig(tpTx.get_bytes())
                tpSignature = tp_auth_data.signature

                if tpSide == ORDER_SIDE_BUY:
                    tpHuman_cost = DECIMAL_CONTEXT_ROUND_UP.multiply(
                        decimal.Decimal(tpSize),
                        decimal.Decimal(tpPrice)
                    )
                    tpFee = DECIMAL_CONTEXT_ROUND_UP.multiply(tpHuman_cost, decimal.Decimal(takerFeeRate))
                else:
                    tpHuman_cost = DECIMAL_CONTEXT_ROUND_DOWN.multiply(
                        decimal.Decimal(tpSize),
                        decimal.Decimal(tpPrice)
                    )
                    tpFee = DECIMAL_CONTEXT_ROUND_DOWN.multiply(tpHuman_cost, decimal.Decimal(takerFeeRate))

                tp_limit_fee_rounded = DECIMAL_CONTEXT_ROUND_UP.quantize(
                    decimal.Decimal(tpFee),
                    decimal.Decimal(currency.get('showStep')), )

        order = {
            'symbol': symbol,
            'side': side,
            'type': type,
            'timeInForce': timeInForce,
            'size': size,
            'price': price,
            'limitFee': str(limit_fee_rounded),
            'expiration': timestampSeconds,
            'triggerPrice': triggerPrice,
            'triggerPriceType': triggerPriceType,
            'trailingPercent': trailingPercent,
            'clientId': clientId,
            'signature': signature,
            'reduceOnly': reduceOnly,
            'isPositionTpsl': isPositionTpsl,
            'isOpenTpslOrder': isOpenTpslOrder,
            'isSetOpenSl': isSetOpenSl,
            'isSetOpenTp': isSetOpenTp,
            'slClientOrderId': slClientId,
            'slPrice': slPrice,
            'slSide': slSide,
            'slSize': slSize,
            'slTriggerPrice': slTriggerPrice,
            'slTriggerPriceType': slTriggerPriceType,
            'slExpiration': slExpiration,
            'slLimitFee': str(sl_limit_fee_rounded),
            'slSignature': slSignature,
            'tpClientOrderId': tpClientId,
            'tpPrice': tpPrice,
            'tpSide': tpSide,
            'tpSize': tpSize,
            'tpTriggerPrice': tpTriggerPrice,
            'tpTriggerPriceType': tpTriggerPriceType,
            'tpExpiration': tpExpiration,
            'tpLimitFee': str(tp_limit_fee_rounded),
            'tpSignature': tpSignature,
            'sourceFlag': sourceFlag,
            'brokerId':brokerId,
        }

        path = URL_SUFFIX + "/v3/order"
        return self._post(
            endpoint=path,
            data=order
        )

    def create_withdrawal_v3(self,
                          amount,
                          toChainId,
                          asset,
                          nonce=None,
                          l2SourceTokenId=None,
                          l1TargetTokenId=None,
                          zkAccountId=None,
                          subAccountId=None,
                          fee='0',
                          clientId=None,
                          timestampSeconds=None,
                          ethAddress=None,
                          isFastWithdraw=False,
                          signature=None, ):

        clientId = clientId or random_client_id()
        if not self.zk_seeds:
            raise Exception(
                'No signature provided and client was not ' +
                'initialized with zk_seeds'
            )

        timestampSeconds =  int(timestampSeconds or int(time.time()))


        ethAddress = ethAddress or self.accountV3.get('ethereumAddress')

        zkAccountId = zkAccountId or self.accountV3.get('spotAccount').get('zkAccountId')

        subAccountId = subAccountId or self.accountV3.get('spotAccount').get('defaultSubAccountId')

        nonce = nonce or self.accountV3.get('spotAccount').get('subAccounts')[0].get('nonce')

        l2Key = self.zk_l2Key or self.accountV3.get('l2Key')
        if not ethAddress:
            raise Exception(
                'No ethAddress provided' +
                'please call get_user()'
            )


        if not self.configV3:
            raise Exception(
                'No config provided' +
                'please call configs_v3()'
            )

        currency = {}

        for k, v in enumerate(self.configV3.get('spotConfig').get('assets')):
            if v.get('token') == asset:
                currency = v

        l2SourceTokenId = l2SourceTokenId or currency.get('tokenId')
        l1TargetTokenId = l1TargetTokenId or currency.get('tokenId')

        withdraw_fee_ratio = 0
        if isFastWithdraw == True:
            withdraw_fee_ratio = decimal.Decimal(fee) * decimal.Decimal(10000) / decimal.Decimal(amount)
            withdraw_fee_ratio = withdraw_fee_ratio.quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)
            fee = (withdraw_fee_ratio * decimal.Decimal(amount) / 10000).__str__()


        amountStr = (decimal.Decimal(amount) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)
        #feeStr = (decimal.Decimal(fee) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)

        builder = sdk.WithdrawBuilder(
            int(zkAccountId),  int(subAccountId), int(toChainId),ethAddress, int(l2SourceTokenId),
            int(l1TargetTokenId), amountStr.__str__(), None, '0', int(nonce),  int(withdraw_fee_ratio),  False, int(timestampSeconds)
        )
        tx = sdk.Withdraw(builder)
        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)

        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature

        withdraw = {
            'amount': amount,
            'timestamp': int(timestampSeconds),
            'ethAddress': ethAddress,
            'clientWithdrawId': clientId,
            'signature': signature,
            'zkAccountId': zkAccountId,
            'subAccountId': subAccountId,
            'l2Key': l2Key,
            'fee': str(fee),
            'toChainId': toChainId,
            'l2SourceTokenId': l2SourceTokenId,
            'l1TargetTokenId': l1TargetTokenId,
            'isFastWithdraw': isFastWithdraw,
            'nonce': nonce,
        }

        path = URL_SUFFIX + "/v3/withdrawal"
        return self._post(
            endpoint=path,
            data=withdraw
        )

    def create_transfer_out_v3(self,
                             amount,
                             asset,
                             nonce=None,
                             tokenId=None,
                             zkAccountId=None,
                             subAccountId=None,
                             fee='0',
                             clientId=None,
                             timestampSeconds=None,
                             receiverAccountId=None,
                             receiverZkAccountId=None,
                             receiverSubAccountId=None,
                             receiverAddress=None,
                             signature=None, ):

        clientId = clientId or random_client_id()
        if not self.zk_seeds:
            raise Exception(
                'No signature provided and client was not ' +
                'initialized with zk_seeds'
            )

        timestampSeconds =  int(timestampSeconds or int(time.time()))
        zkAccountId = zkAccountId or self.accountV3.get('spotAccount').get('zkAccountId')
        subAccountId = subAccountId or self.accountV3.get('spotAccount').get('defaultSubAccountId')
        nonce = nonce or self.accountV3.get('spotAccount').get('subAccounts')[0].get('nonce')

        receiverAddress = receiverAddress or self.configV3.get('spotConfig').get('global').get('contractAssetPoolEthAddress')
        receiverZkAccountId = receiverZkAccountId or self.configV3.get('spotConfig').get('global').get('contractAssetPoolZkAccountId')
        receiverSubAccountId = receiverSubAccountId or self.configV3.get('spotConfig').get('global').get('contractAssetPoolSubAccount')
        receiverAccountId = receiverAccountId or self.configV3.get('spotConfig').get('global').get('contractAssetPoolAccountId')

        if not self.configV3:
            raise Exception(
                'No config provided' +
                'please call configs_v3()'
            )

        currency = {}

        for k, v in enumerate(self.configV3.get('spotConfig').get('assets')):
            if v.get('token') == asset:
                currency = v

        tokenId = tokenId or currency.get('tokenId')

        amountStr = (decimal.Decimal(amount) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)

        builder = sdk.TransferBuilder(
            int(zkAccountId),  receiverAddress, int(subAccountId), int(receiverSubAccountId), int(tokenId), amountStr.__str__(), '0', int(nonce),  int(timestampSeconds)
        )

        tx = sdk.Transfer(builder)
        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)

        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature

        transferData = {
            'amount': amount,
            'timestamp': int(timestampSeconds),
            'clientTransferId': clientId,
            'signature': signature,
            'zkAccountId': zkAccountId,
            'subAccountId': subAccountId,
            'fee': str(fee),
            'token': asset,
            'tokenId': tokenId,
            'receiverAccountId': receiverAccountId,
            'receiverZkAccountId': receiverZkAccountId,
            'receiverSubAccountId': receiverSubAccountId,
            'receiverAddress': receiverAddress,
            'nonce': nonce,
        }

        path = URL_SUFFIX + "/v3/transfer-out"
        return self._post(
            endpoint=path,
            data=transferData
        )

    def create_transfer_out_to_address_v3(self,
                               amount,
                               asset,
                               nonce=None,
                               tokenId=None,
                               zkAccountId=None,
                               subAccountId=None,
                               fee='0',
                               clientId=None,
                               timestampSeconds=None,
                               receiverAccountId=None,
                               receiverZkAccountId=None,
                               receiverSubAccountId=None,
                               receiverAddress=None,
                               signature=None, ):

        clientId = clientId or random_client_id()
        if not self.zk_seeds:
            raise Exception(
                'No signature provided and client was not ' +
                'initialized with zk_seeds'
            )

        timestampSeconds =  int(timestampSeconds or int(time.time()))
        zkAccountId = zkAccountId or self.accountV3.get('spotAccount').get('zkAccountId')
        subAccountId = subAccountId or self.accountV3.get('spotAccount').get('defaultSubAccountId')
        nonce = nonce or self.accountV3.get('spotAccount').get('subAccounts')[0].get('nonce')

        receiverSubAccountId = receiverSubAccountId or 0

        if not self.configV3:
            raise Exception(
                'No config provided' +
                'please call configs_v3()'
            )

        currency = {}

        for k, v in enumerate(self.configV3.get('spotConfig').get('assets')):
            if v.get('token') == asset:
                currency = v

        tokenId = tokenId or currency.get('tokenId')

        amountStr = (decimal.Decimal(amount) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)

        builder = sdk.TransferBuilder(
            int(zkAccountId),  receiverAddress, int(subAccountId), int(receiverSubAccountId), int(tokenId), amountStr.__str__(), '0', int(nonce),  int(timestampSeconds)
        )

        tx = sdk.Transfer(builder)
        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)

        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature

        transferData = {
            'amount': amount,
            'timestamp': int(timestampSeconds),
            'clientTransferId': clientId,
            'signature': signature,
            'zkAccountId': zkAccountId,
            'subAccountId': subAccountId,
            'fee': str(fee),
            'token': asset,
            'tokenId': tokenId,
            'receiverAccountId': receiverAccountId,
            'receiverZkAccountId': receiverZkAccountId,
            'receiverSubAccountId': receiverSubAccountId,
            'receiverAddress': receiverAddress,
            'nonce': nonce,
        }

        path = URL_SUFFIX + "/v3/transfer-out-to-address"
        return self._post(
            endpoint=path,
            data=transferData
        )

    def create_contract_transfer_out_v3(self,
                                        amount,
                                        asset,
                                        nonce=None,
                                        tokenId=None,
                                        zkAccountId=None,
                                        subAccountId=None,
                                        ethAddress=None,
                                        clientId=None,
                                        timestampSeconds=None,
                                        accountId=None,
                                        receiverAccountId=None,
                                        receiverZkAccountId=None,
                                        receiverSubAccountId=None,
                                        receiverAddress=None,
                                        signature=None, ):

        clientId = clientId or random_client_id()
        if not self.zk_seeds:
            raise Exception(
             'No signature provided and client was not ' +
             'initialized with zk_seeds'
            )

        timestampSeconds =  timestampSeconds or int(time.time())
        timestampSeconds = int(timestampSeconds + 3600 * 24 * 28)
        accountId = accountId or self.accountV3.get('id')
        subAccountId = subAccountId or self.accountV3.get('spotAccount').get('defaultSubAccountId')

        ethAddress = ethAddress or self.accountV3.get('ethereumAddress')

        receiverSubAccountId = receiverSubAccountId or 0

        message = hashlib.sha256()
        message.update(clientId.encode())  # Encode as UTF-8.
        nonceHash = message.hexdigest()
        nonceInt = int(nonceHash, 16)
        maxUint32 = np.iinfo(np.uint32).max


        nonce = nonceInt % maxUint32
        accountId = int(accountId, 10) % maxUint32

        if not self.configV3:
            raise Exception(
               'No config provided' +
               'please call configs_v3()'
            )

        currency = {}

        for k, v in enumerate(self.configV3.get('contractConfig').get('assets')):
            if v.get('token') == asset:
                currency = v

        tokenId = tokenId or currency.get('tokenId')

        amountStr = (decimal.Decimal(amount) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)

        builder = sdk.TransferBuilder(
            int(accountId),  ethAddress, int(subAccountId), int(receiverSubAccountId),  int(tokenId), amountStr.__str__(), '0', int(nonce),  int(timestampSeconds)
        )

        tx = sdk.Transfer(builder)
        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)

        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature

        transferData = {
            'amount': amount,
            'expireTime': timestampSeconds,
            'clientWithdrawId': clientId,
            'signature': signature,
            'token': asset,
            'ethAddress': ethAddress,
        }

        path = URL_SUFFIX + "/v3/contract-transfer-out"
        return self._post(
            endpoint=path,
            data=transferData
        )

    def create_contract_transfer_to_address_v3(self,
                                        amount,
                                        asset,
                                        receiverL2Key,
                                        receiverAccountId,
                                        receiverAddress,
                                        receiverSubAccountId=None,
                                        nonce=None,
                                        tokenId=None,
                                        subAccountId=None,
                                        clientId=None,
                                        timestampSeconds=None,
                                        accountId=None,
                                        signature=None, ):

        clientId = clientId or random_client_id()
        if not self.zk_seeds:
            raise Exception(
                'No signature provided and client was not ' +
                'initialized with zk_seeds'
            )

        timestampSeconds =  timestampSeconds or int(time.time())
        timestampSeconds = int(timestampSeconds + 3600 * 24 * 28)
        accountId = accountId or self.accountV3.get('id')
        subAccountId = subAccountId or self.accountV3.get('spotAccount').get('defaultSubAccountId')
        receiverSubAccountId = receiverSubAccountId or 0

        message = hashlib.sha256()
        message.update(clientId.encode())  # Encode as UTF-8.
        nonceHash = message.hexdigest()
        nonceInt = int(nonceHash, 16)
        maxUint32 = np.iinfo(np.uint32).max


        nonce = nonceInt % maxUint32
        accountId = int(accountId, 10) % maxUint32

        if not self.configV3:
            raise Exception(
                'No config provided' +
                'please call configs_v3()'
            )

        currency = {}

        for k, v in enumerate(self.configV3.get('contractConfig').get('assets')):
            if v.get('token') == asset:
                currency = v

        tokenId = tokenId or currency.get('tokenId')

        amountStr = (decimal.Decimal(amount) * decimal.Decimal(10) ** decimal.Decimal(currency.get('decimals'))).quantize(decimal.Decimal(0), rounding=decimal.ROUND_UP)

        builder = sdk.TransferBuilder(
            int(accountId),  receiverAddress, int(subAccountId), int(receiverSubAccountId),  int(tokenId), amountStr.__str__(), '0', int(nonce),  int(timestampSeconds)
        )

        tx = sdk.Transfer(builder)
        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)

        auth_data = signerSeed.sign_musig(tx.get_bytes())
        signature = auth_data.signature

        transferData = {
            'amount': amount,
            'expireTime': timestampSeconds,
            'clientTransferId': clientId,
            'signature': signature,
            'token': asset,
            'receiverAccountId': receiverAccountId,
            'receiverL2Key': receiverL2Key,
        }

        path = URL_SUFFIX + "/v3/contract-transfer-to"
        return self._post(
            endpoint=path,
            data=transferData
        )

    def create_manual_repayment_v3(self,
                                   repaymentTokens,
                                   poolRepaymentTokens,
                                   clientId=None,
                                   timestampSeconds=None,
                                   accountId=None,
                                   signature=None, ):

        clientId = clientId or random_client_id()
        if not self.zk_seeds:
            raise Exception(
                'No signature provided and client was not ' +
                'initialized with zk_seeds'
            )

        timestampSeconds =  timestampSeconds or int(time.time())
        timestampSeconds = int(timestampSeconds + 3600 * 24 * 28)
        accountId = accountId or self.accountV3.get('id')

        msgHashString ="accountId=" + accountId + "&clientId=" + clientId + "&expireTime=" + str(timestampSeconds) + "&repaymentTokens=" + repaymentTokens

        message = hashlib.sha256()
        message.update(msgHashString.encode())  # Encode as UTF-8.
        msgHash = message.digest()

        EC_ORDER = '3618502788666131213697322783095070105526743751716087489154079457884512865583';

        bn1 = int(msgHash.hex(), 16)
        bn2 = int (EC_ORDER, 10)
        signMsg = hex(bn1.__mod__(bn2))

        seedsByte = bytes.fromhex(self.zk_seeds.removeprefix('0x') )
        signerSeed = sdk.ZkLinkSigner().new_from_seed(seedsByte)
        signatureData = signerSeed.sign_musig(signMsg.removeprefix('0x').encode())
        signature = signatureData.signature

        repaymentData = {
            'repaymentTokens': repaymentTokens,
            'expireTime': timestampSeconds,
            'clientId': clientId,
            'signature': signature,
            'poolRepaymentTokens': poolRepaymentTokens,
        }

        path = URL_SUFFIX + "/v3/manual-create-repayment"
        return self._post(
            endpoint=path,
            data=repaymentData
        )
